#!/usr/local/bin/gawk -f

# dpkglist.awk
# Neal Burns, 2011
# nnneal *at* gmail *dot com

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

function show_help() {
  print "Usage: dpkglist.awk [<namepattern> | <field>=<regexp> | <variable>=<value> | files=<filelist>] ..."
  print "       dpkglist.awk -- --help (show longer help)"
  print "       dpkglist.awk -e <script>"
  print ""
  print "Package name is the default field for matching. All matching is performed case-insensitively by default."
  print "Installed packages that match, if any, are listed first and are colored cyan."
  print ""
  print "Examples:"
  print "   Show all packages with names containing libxml2:"
  print "       dpkglist.awk libxml2"
  print "   Show all packages with descriptions containing xslt:"
  print "       dpkglist.awk d=xslt"
  print "   Show all packages with descriptions containing both xslt and python, and names containing dbg:"
  print "       dpkglist.awk d=xslt d=python dbg"
  print "   Show only the name and description columns, with 30 characters for the name, and the remainder for the description:"
  print "       dpkglist.awk libxml2 cols=\"Package Description\" widths=\"30\""
  print "   Show name and description only, on separate lines, with field headers prepended, no limit on width, and newlines preserved (with bash as shell):"
  print "       dpkglist.awk libxml2 cols=\"Package Description\" widths=\"0 0\" keepnewlines=1 ofs=$'\\n' ors=$'\\n\\n' headertype=pre"
}

function show_long_help() {
  show_help()
  print ""
  print "Common fields are (shorted form): name (n), section (s), version (v), description (d), origin (o), depends (dep)"
  print "Any additional field can be specified using its full name, ex: dpkglist.awk Website='sourceforge\\.net'"
  print "Full names of fields are always capitalized. This allows them to be distinguished from variable names."
  print ""
  print "Package list files may be specified on the command line by setting the files variable. This variable should contain a"
  print "space-separated list of file names. If none are explicitly specified, the default is /var/lib/apt/lists/*_Packages, and"
  print "statuses are taken from /var/lib/dpkg/status."
  print ""
  print "Configuration variables:"
  print "* ic: false = consider case in pattern matches (default is true)"
  print "* cols is a space-delimited list of field names that specifies the output columns. The field names are as in the apt cache."
  print "* Note that the package name field is called Package, not name, when specified in the cols variable."
  print "* Ex: cols='Section Package Version Description' (this is the default)"
  print "* widths is a space-delimited list of integers that specify the output column widths. If width is negative, the column is"
  print "* right-justified."
  print "* If a column width is specified as 0, the field is printed as-is: no truncation or padding occurs. If the last column width is"
  print "* omitted, the width of the last column will be calculated automatically so as to fill all remaining space on the line."
  print "* headertype: 'none' = suppress header, 'top' (default) = print headers first, 'pre' = prepend to each field,"
  print "* noprint: true = suppress automatic printing if the pattern variables are satisfied."
  print "* nocolor: true = do not output vt100 color codes."
  print "* defaultwidth: sets the widths of fields that were not specified and have no internal default."
  print "* ofs and ors: set the output field separator (OFS) and output record separator (ORS), respectively"
  print "* ORS and OFS will be used to separate the output records (packages) and output fields."
  print "* keepnewlines: true = leave newlines intact. Newlines occur primarily in the Description field."
  print "* noprefilter: true = disable the automatic pre-filter (only useful to avoid the bug mentioned below)"
  print "Note: To set a variable to false, set it to 0 or the empty string. Any other value is interpreted as true."
  print ""
  print "Informational variables: pkgmatch (nonzero if the pattern variables have been satisfied; for use in scripts)."
  print ""
  print "BUGS: The pre-filter doesn't know how to handle escaped anchors in field patterns, so, for instance, it is impossible to search"
  print "for a literal ^ or $ character while the pre-filter is enabled. Other small annoyances."
  print ""
  print "Homepage: http://sourceforge.net/projects/dpkglistawk/"
}

BEGIN {
  if(ARGC == 1) {
    show_help()
    exit
  }

  # IGNORECASE is the default behavior
  IGNORECASE = 1

  # add the input file to the end of the command line if no other files are specified
  for(arg in ARGV) {
    if(arg == 0) {
      continue
    }
    if(ARGV[arg] ~ /--help|-h|-\?/) {
      show_long_help()
      exit
    }
    if(ARGV[arg] !~ "=") {
      add_search_pattern("Package", ARGV[arg])
      ARGV[arg] = ""
    }
    else {
      if(get_command_line_variable(ARGV[arg])) {
        ARGV[arg] = ""
      }
    }
  }
  if(files) {
    split(files, file_arr, " ")
    for(i in file_arr) {
      ARGV[ARGC++] = file_arr[i]
    }
  }
  else {
    ARGV[ARGC++] = "/var/lib/dpkg/status"
    if(!statusonly) {
      command = "ls /var/lib/apt/lists/*_Packages"
      while ((command | getline file) > 0) {
        ARGV[ARGC++] = file
      }
      close(command)
    }
  }

  RS = "\n\n"
  FS = "\n\\<"
}

function add_search_pattern(field, pattern) {
  if(!(field in search_patterns)) {
    search_patterns[field][0] = pattern
  }
  else {
    search_patterns[field][length(search_patterns[field])] = pattern
  }
}

# returns true if a recognized variable was parsed
function get_command_line_variable(arg,       a, temp) {
  # match: string, regex, array
  match(arg, /^(\w+)=(.*)/, a)

  # custom field pattern specification (variable name starts with an uppercase letter)
  temp = IGNORECASE
  IGNORECASE = 0
  if(a[1] ~ /^[[:upper:]]/) {
    add_search_pattern(a[1], a[2])
    return 1
  }
  IGNORECASE = temp

  # handle the regex variables explicity, because it eliminates the need to double-escape special characters
  switch(a[1]) {
  case "files":
    files = files " " gensub(/^files=/, "", "", ARGV[arg])
    break;
  case "name":
  case "n":
    add_search_pattern("Package", a[2])
    break;
  case "section":
  case "s":
    add_search_pattern("Section", a[2])
    break
  case "version":
  case "v":
    add_search_pattern("Version", a[2])
    break
  case "description":
  case "d":
    add_search_pattern("Description", a[2])
    break
  case "origin":
  case "o":
    add_search_pattern("Origin", a[2])
    break
  case "depends":
  case "dep":
    add_search_pattern("Depends", a[2])
    break
  case "status":
  case "st":
    add_search_pattern("Status", a[2])
    statusonly = 1
    break
  case "ic":
    if(!a[2]) {
      IGNORECASE = 0
    }
    break
  case "ofs":
      OFS = a[2]
      break
  case "ors":
      ORS = a[2]
      break
  case "all":
    statusonly = !a[2]
    break
  default:
    return 0
  }
  return 1
}

# get patterns from the command line
NR==1 {
  if(!noprefilter) {
    make_search_pattern_prefilter()
  }
}

function make_search_pattern_prefilter(fld, prefilter_pattern, prefilter_fs, i) {
  prefilter_fs = gensub("\n", "\\\\n", "g", FS);

  for(fld in search_patterns) {
    for(i in search_patterns[fld]) {
      prefilter_pattern = search_patterns[fld][i]

      # this is flawed: it doesn't consider whether the anchors are backslash-escaped.
      # I think the way to handle it is to use simple string functions, rather than regex replacements
      # figuring out how to write the regexs properly with all of the multiple escaping is nearly impossible
      gsub(/\$|\\'/, "(" prefilter_fs "|$)", prefilter_pattern)
      gsub(/\^|\\`/, "(" prefilter_fs "|^)" fld ": ", prefilter_pattern)
      prefilter_pattern = sprintf("((%s|^)%s: ([^\\n]*|\\n )*)?(%s)", prefilter_fs, fld, prefilter_pattern)

      if(fld in search_pattern_prefilter) {
        search_pattern_prefilter[fld][length(search_pattern_prefilter[fld])] = prefilter_pattern
      }
      else {
        search_pattern_prefilter[fld][0] = prefilter_pattern
      }
    }
  }
}

# process the record into fields
# no parameters: i and a are local variables
function get_fields(        i, a) {
  if(got_fields == NR) {
    # already got fields for this record
    print "Already got fields!"
    return
  }
  got_fields = NR

  delete f
  delete fi
  delete field
  name = section = version = description = origin = depends = status = ""
  namei = sectioni = versioni = descriptioni = origini = dependsi = statusi = 0
  field[0] = $0
  for(i=1; i<=NF; i++) {
    field[i] = $i

    match($i, /([^:]*): (.*)/, a);

    if(!keepnewlines) {
      gsub(/\n/, "", a[2]) # get rid of newlines in the field value
    }

    f[a[1]] = a[2]
    fi[a[1]] = i

    switch(a[1]) {
      case "Package":
        namei = i
        name = a[2]
        break
      case "Section":
        sectioni = i
        section = a[2]
        break
      case "Version":
        versioni = i
        version = a[2]
        break
      case "Description":
        descriptioni = i
        description = a[2]
        break
      case "Origin":
        origini = i
        origin = a[2]
        break
      case "Depends":
        dependsi = i
        depends = a[2]
        break
      case "Status":
        statusi = i
        status = a[2]
        break
    }
  }
  installed = status ~ /installed/
}

BEGIN {
  # this gets what would be $COLUMNS in bash
  "tput cols" | getline columns
  close("tput cols")
}

# set $0 to formatted output
NR == 1 {
  if(!headertype) {
    # headers on top is the default
    headertype = "top"
  }

  if(cols) {
    split(cols, display_cols, " ");
  }
  else {
    display_cols[1] = "Section"
    display_cols[2] = "Package"
    display_cols[3] = "Version"
    display_cols[4] = "Description"
  }
  NUMCOLS = length(display_cols)
  if(widths) {
    split(widths, display_widths, " ");
  }
  else {
    for(i=1; i<NUMCOLS; i++) {
      switch(display_cols[i]) {
        case "Section":
          display_widths[i] = 10
          break
        case "Package":
          display_widths[i] = 20
          break
        case "Version":
          display_widths[i] = -15 # negative width means column should be right-justified
          break
        default:
          if(defaultwidth) {
            display_widths[i] = defaultwidth
          }
          else {
            display_widths[i] = 20 # arbitrary default
          }
      }
    }
    # if last width is omitted, it will be calculated based on the widths of the prior columns
  }

  header = "";
  for(i=1; i<=NUMCOLS-1; i++) {
    if(display_widths[i] > 0) {
      header = header sprintf("%-*.*s", display_widths[i], display_widths[i], display_cols[i]) OFS
    }
    else if(display_widths[i] < 0) {
      header = header sprintf("%*.*s", -display_widths[i], -display_widths[i], display_cols[i]) OFS
    }
    else {
      header = header sprintf("%s", display_cols[i]) OFS
    }
  }

  # if last column is unset, calculate a value to maximize it
  if(!(NUMCOLS in display_widths)) {
    initialwidth = index(header ORS, "\n") - 1;
    display_widths[NUMCOLS] = columns - initialwidth
    if(display_widths[NUMCOLS] <= 0) {
      print "Fields exceed terminal width!"
      exit 1
    }
  }

  if(display_widths[NUMCOLS] > 0) {
    header = header sprintf("%-*.*s", display_widths[NUMCOLS], display_widths[NUMCOLS], display_cols[NUMCOLS])
  }
  else if(display_widths[NUMCOLS] < 0) {
    header = header sprintf("%*.*s", display_widths[NUMCOLS], display_widths[NUMCOLS], display_cols[NUMCOLS])
  }
  else {
    header = header sprintf("%s", display_cols[NUMCOLS])
  }

  if(headertype == "top" && !noprint) {
    if(!nocolor) {
      printf("\x1b[%d;1m%s\x1b[0m%s", 34, header, ORS);
    }
    else {
      printf("%s%s", header, ORS)
    }
  }
}

# set pkgmatch as appropriate
!length(search_patterns) || noprefilter || do_prefilter() {
  get_fields()

  pkgmatch = do_search_patterns()

  if(!pkgmatch || !NUMCOLS) {
    next
  }


  format_fields();

  if(sort) {
    lines[name]["status"] = status
    for(i=1; i<=NF; i++) {
      lines[name][i] = $i
    }
  }
  else if(!noprint) {
    print
  }
}

# called internally by gawk. i, tmp1, and tmp2 are local variables
function sort_compare(i1, v1, i2, v2,      i, tmp1, tmp2) {
  # sort on the column in sort
  for(i=1; i<=num_sort_columns; i++) {
    # remove any leading whitespace to make the sort more useful if the field is right-aligned
    tmp1 = gensub(/^\s*/, "", 1, tolower(v1[sort_columns[i]]))
    tmp2 = gensub(/^\s*/, "", 1, tolower(v2[sort_columns[i]]))
    if (tmp1 < tmp2)
      return -1
    if (tmp1 > tmp2)
      return 1
  }
  return 0
}

END {
  if(sort) {
    num_sort_columns = split(sort, sort_columns, " ")
    nlines = asort(lines, lines, "sort_compare")
    for(line=1; line<=nlines; line++) {
      NF = 0

      installed = lines[line]["status"] ~ /installed/

      for(i=1; i<=NUMCOLS; i++) {
        $i = lines[line][i]

        if(!installed && display_cols[i] in search_patterns) {
          for(j in search_patterns[display_cols[i]]) {
            add_color(search_patterns[display_cols[i]][j], i)
          }
        }
        else if(installed) {
          add_color(".*", i, 36, 1)
        }
      }

      print
    }
  }
}

# no parameters: i is a local variable
function format_fields(         i) {
  NF = 0
  for(i=1; i<=NUMCOLS; i++) {
    if(display_widths[i] > 0) {
      $(++NF) = sprintf("%-*.*s", display_widths[i], display_widths[i], f[display_cols[i]])
      if(!installed && !sort && display_cols[i] in search_patterns) {
        for(j in search_patterns[display_cols[i]]) {
          add_color(search_patterns[display_cols[i]][j], NF)
        }
      }
    }
    else if(display_widths[i] < 0) {
      $(++NF) = sprintf("%*.*s", -display_widths[i], -display_widths[i], f[display_cols[i]])
      if(!installed && !sort && display_cols[i] in search_patterns) {
        for(j in search_patterns[display_cols[i]]) {
          add_color(search_patterns[display_cols[i]][j], NF)
        }
      }
    }
    else {
      $(++NF) = sprintf("%s", f[display_cols[i]])
    }

    if(headertype == "pre") {
      $NF = display_cols[i] ": " $NF
    }
  }

  if(!sort && installed) {
    for(i=1; i<=NF; i++) {
      # turn installed packages cyan
      add_color(".*", i, 36, 1)
    }
  }
}

# no parameters: i is a local variable
function do_prefilter(        i) {
  # remove duplicates
  if($1 in packages_seen) {
    return 0
  }
  packages_seen[$1] = 1

  for(pat in search_pattern_prefilter) {
    for(i in search_pattern_prefilter[pat]) {
      if($0 !~ search_pattern_prefilter[pat][i]) {
        return 0
      }
    }
  }
  return 1
}

# no parameters: i is a local variable
function do_search_patterns(        i) {
  for(pat in search_patterns) {
    if(!(pat in f)) {
      return 0
    }
    for(i in search_patterns[pat]) {
      if(f[pat] !~ search_patterns[pat][i]) {
        return 0
      }
    }
  }
  return 1
}

function add_color(expr, fieldnum, color, nohighlight) {
  if(!nocolor) {
    if(!color)
      color = 32 # default to green if the color is not specified
    if(!nohighlight) {
      gsub(expr, "\x1b[" color ";1m&\x1b[0m", $fieldnum);
    }
    else {
      gsub(expr, "\x1b[" color "m&\x1b[0m", $fieldnum);
    }
  }
}

function add_to_pattern(pattern, to_add)
{
  if(pattern ~ /^\s*$/) {
    return to_add
  }
  else {
    return pattern "|" to_add
  }
}

